"""
    Helper to publish this Project to PyPi
"""

import subprocess
from pathlib import Path

from bx_py_utils.path import assert_is_file
from cli_base.cli_tools.subprocess_utils import verbose_check_call
from manageprojects.utilities.publish import publish_package

import bx_django_utils


def publish():
    """
    Publish to PyPi
    Call this via:
        $ .venv/bin/publish
    """
    package_path = Path(bx_django_utils.__file__).parent.parent
    assert_is_file(package_path / 'pyproject.toml')

    subprocess.check_call(['pipenv', 'check'])

    verbose_check_call('make', 'test')  # don't publish if tests fail
    verbose_check_call('make', 'fix-code-style')  # don't publish if code style wrong

    publish_package(
        module=bx_django_utils,
        package_path=package_path,
    )


if __name__ == '__main__':
    publish()



import datetime
import logging
import time
from collections.abc import Callable, Iterable, Iterator
from contextlib import contextmanager

from django.core.cache import cache

from bx_django_utils.feature_flags.exceptions import NotUniqueFlag
from bx_django_utils.feature_flags.models import FeatureFlagModel
from bx_django_utils.feature_flags.state import State
from bx_django_utils.feature_flags.utils import validate_cache_key
from bx_django_utils.models.manipulate import create_or_update2

logger = logging.getLogger(__name__)

DEFAULT_DURATION = datetime.timedelta(seconds=0)


class FeatureFlag:
    """
    A feature flag that persistent the state into django cache/database.
    """

    registry = {}

    def __init__(
        self,
        *,
        cache_key: str,
        human_name: str,
        initial_enabled: bool,
        description: str | None = None,
        cache_key_prefix: str = 'feature-flags',
        cache_duration: datetime.timedelta = DEFAULT_DURATION,
    ):
        """
        :param cache_duration: how long the state of the flag should be cached in-process
        """
        self.human_name = human_name
        self.description = description

        validate_cache_key(cache_key)
        validate_cache_key(cache_key_prefix)
        self.name = cache_key
        self.cache_key = f'{cache_key_prefix}-{cache_key}'
        validate_cache_key(self.cache_key)  # Double check ;)

        if self.cache_key in self.registry:
            raise NotUniqueFlag(f'Cache key {self.cache_key !r} already registered!')

        self.registry[self.cache_key] = self

        if initial_enabled:
            self.initial_state = State.ENABLED
        else:
            self.initial_state = State.DISABLED

        if cache_duration:
            self._cache_duration: datetime.timedelta = cache_duration
            self._cache_time_func: Callable[[], float] = time.monotonic
            self._cache_from: float = self._cache_time_func()
            self._cache_value: State = self.initial_state

    def enable(self) -> bool:
        return self.set_state(new_state=State.ENABLED)

    def disable(self) -> bool:
        return self.set_state(new_state=State.DISABLED)

    def set_state(self, new_state: State) -> bool:
        """store state to database + cache and return it"""
        assert isinstance(
            new_state, State
        ), f'Given {new_state!r} (type: {type(new_state).__name__}) is not a State object!'

        state = self._add2cache(state_value=new_state.value)
        create_or_update2(
            ModelClass=FeatureFlagModel,
            lookup={'cache_key': self.cache_key},
            state=new_state,
        )
        return state

    def _add2cache(self, state_value: int) -> bool:
        """store state to cache and return it"""
        assert state_value in (0, 1), f'Unknown cache value: {state_value!r}'
        cache.set(self.cache_key, state_value, timeout=None)  # set forever
        return bool(state_value)

    def reset(self) -> None:
        FeatureFlagModel.objects.filter(cache_key=self.cache_key).delete()
        cache.delete(self.cache_key)

    @property
    def is_enabled(self) -> bool:
        if not hasattr(self, '_cache_duration'):  # caching is disabled
            return self._compute_is_enabled()

        elapsed = self._cache_time_func() - self._cache_from

        # cache is still valid
        if elapsed <= self._cache_duration.total_seconds():
            return bool(self._cache_value.value)

        # cache is invalid -> recompute
        state = self._compute_is_enabled()
        self._cache_from = self._cache_time_func()
        self._cache_value = State(state)
        return state

    def _compute_is_enabled(self) -> bool:
        try:
            raw_value = cache.get(self.cache_key)
        except Exception as err:
            # Will only happen, if cache backend is down
            logger.exception('Get cache key %r failed: %s', self.cache_key, err)
            raw_value = None  # Use the initial state

        if raw_value is None:
            # Not set, yet or cache was cleared -> Try to get the state from database
            instance = (
                FeatureFlagModel.objects.filter(cache_key=self.cache_key).values('state').first()
            )
            if instance:
                # We found the flag in database -> store state in the cache
                # and return the current state
                state_value = instance['state']
                return self._add2cache(state_value=state_value)
            else:
                # It's not in the cache and database -> store initial state in cache and database
                # and return the current state
                return self.set_state(new_state=self.initial_state)
        else:
            assert raw_value in (0, 1), f'Unknown cache value: {raw_value!r}'
            return bool(raw_value)

    @property
    def is_disabled(self) -> bool:
        return not self.is_enabled

    @contextmanager
    def temp_enable(self) -> Iterator[None]:
        """
        Temporarily enable the feature flag.
        """
        was_enabled = self.is_enabled
        if not was_enabled:
            self.enable()

        yield

        if not was_enabled:
            self.disable()

    @contextmanager
    def temp_disable(self) -> Iterator[None]:
        """
        Temporarily disable the feature flag.
        """
        was_disabled = self.is_disabled
        if not was_disabled:
            self.disable()

        yield

        if not was_disabled:
            self.enable()

    @property
    def state(self):
        if self.is_enabled:
            return State.ENABLED
        else:
            return State.DISABLED

    @property
    def opposite_state(self):
        """
        returns the opposite of current "state".
        Usefull to toggle the state via set_state()
        """
        if self.is_enabled:
            return State.DISABLED
        else:
            return State.ENABLED

    @classmethod
    def values(cls) -> Iterable["FeatureFlag"]:
        yield from cls.registry.values()

    @classmethod
    def get_by_cache_key(cls, cache_key) -> "FeatureFlag":
        return cls.registry[cache_key]

    def __bool__(self):
        return self.is_enabled

    def __str__(self):
        return f'<FeatureFlag {self.name}>'

    def __repr__(self):
        initial = self.initial_state == State.ENABLED
        return f'FeatureFlag(cache_key={self.name!r}, human_name={self.human_name!r}, initial_enabled={initial!r})'



import logging
import re
import zoneinfo

from django.utils import timezone


logger = logging.getLogger(__name__)


class InvalidUserTimeZone(ValueError):
    pass


def validate(raw_time_zone: str, min_length: int, max_length: int, usertimezone_re: str) -> None:
    """
    Validate the UserTimeZone cookie value.
    """
    length = len(raw_time_zone)

    if length > max_length:
        raise InvalidUserTimeZone(f'Expected max length {max_length}, got {length}')

    if length < min_length:
        raise InvalidUserTimeZone(f'Expected min length {min_length}, got {length}')

    if not raw_time_zone.isascii():
        raise InvalidUserTimeZone('Only ASCII chars allowed')

    if not re.fullmatch(usertimezone_re, raw_time_zone):
        raise InvalidUserTimeZone(f'Not match: {raw_time_zone}')


class UserTimezoneMiddleware:
    """
    Activate Timezone by "UserTimeZone" cookie
    """

    MIN_LENGTH = 2  # e.g.: "NZ", "GB", etc.
    MAX_LENGTH = 32  # e.g.: "America/Argentina/ComodRivadavia"
    USERTIMEZONE_RE = r'[a-zA-Z0-9_\-\+/]+'  # e.g.: "Etc/GMT+10", "US/East-Indiana", "Europe/San_Marino"

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if raw_time_zone := request.COOKIES.get('UserTimeZone'):
            self.validate(raw_time_zone)
            try:
                tzinfo = zoneinfo.ZoneInfo(raw_time_zone)
            except zoneinfo.ZoneInfoNotFoundError:
                logger.error('Unknown user time zone %r', raw_time_zone)
            else:
                # Response with user timezone:
                with timezone.override(tzinfo):
                    response = self.get_response(request)
                return response

        # Response with default timezone:
        return self.get_response(request)

    def validate(self, raw_time_zone):
        validate(
            raw_time_zone=raw_time_zone,
            min_length=self.MIN_LENGTH,
            max_length=self.MAX_LENGTH,
            usertimezone_re=self.USERTIMEZONE_RE,
        )



# Generated by Django 4.2.11 on 2024-04-16 06:12

import django.db.models.deletion
from django.db import migrations, models

import bx_django_utils.data_types.gtin.model_fields
import bx_django_utils.data_types.gtin.validators
import bx_django_utils.models.color_field
import bx_django_utils.translation


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='RawTranslatedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('translated', models.JSONField()),
                ('translated_multiline', models.JSONField(blank=True, default=dict)),
                ('not_translated', models.TextField(default='A Default Value')),
            ],
            options={
                'db_table': 'test_app_translatedmodel',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='ColorFieldTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('required_color', bx_django_utils.models.color_field.ColorModelField(max_length=7)),
                (
                    'optional_color',
                    bx_django_utils.models.color_field.ColorModelField(blank=True, max_length=7, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name='ConnectedUniqueSlugModel1',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'translated',
                    bx_django_utils.translation.TranslationField(
                        languages=(('de-de', 'German'), ('en-us', 'US-English'), ('es', 'Spanish')),
                        max_value_length=None,
                        min_value_length=None,
                    ),
                ),
                (
                    'translated_slug',
                    bx_django_utils.translation.TranslationSlugField(
                        additional_uniqueness=(
                            {
                                'app_label': 'test_app',
                                'field_name': 'translated_slug',
                                'model_name': 'ConnectedUniqueSlugModel2',
                            },
                        ),
                        blank=True,
                        languages=(('de-de', 'German'), ('en-us', 'US-English'), ('es', 'Spanish')),
                        max_value_length=None,
                        min_value_length=None,
                        unique=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='ConnectedUniqueSlugModel2',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'translated',
                    bx_django_utils.translation.TranslationField(
                        languages=(('de-de', 'German'), ('en-us', 'US-English'), ('es', 'Spanish')),
                        max_value_length=None,
                        min_value_length=None,
                    ),
                ),
                (
                    'translated_slug',
                    bx_django_utils.translation.TranslationSlugField(
                        additional_uniqueness=(
                            {
                                'app_label': 'test_app',
                                'field_name': 'translated_slug',
                                'model_name': 'ConnectedUniqueSlugModel1',
                            },
                        ),
                        blank=True,
                        languages=(('de-de', 'German'), ('en-us', 'US-English'), ('es', 'Spanish')),
                        max_value_length=None,
                        min_value_length=None,
                        unique=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='GtinFieldTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'default_gtin',
                    bx_django_utils.data_types.gtin.model_fields.GtinModelField(
                        max_length=14,
                        validators=[
                            bx_django_utils.data_types.gtin.validators.GtinValidator(accepted_length=(12, 13, 14))
                        ],
                        verbose_name='GTIN',
                    ),
                ),
                (
                    'all_gtin',
                    bx_django_utils.data_types.gtin.model_fields.GtinModelField(
                        blank=True,
                        max_length=14,
                        validators=[
                            bx_django_utils.data_types.gtin.validators.GtinValidator(accepted_length=(12, 13, 14))
                        ],
                        verbose_name='GTIN',
                    ),
                ),
                (
                    'ean13',
                    bx_django_utils.data_types.gtin.model_fields.GtinModelField(
                        blank=True,
                        max_length=14,
                        null=True,
                        validators=[
                            bx_django_utils.data_types.gtin.validators.GtinValidator(accepted_length=(12, 13, 14))
                        ],
                        verbose_name='GTIN',
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='NonUniqueTranslatedSlugTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'translated',
                    bx_django_utils.translation.TranslationField(
                        languages=(('de-de', 'German'), ('en-us', 'US-English'), ('es', 'Spanish')),
                        max_value_length=None,
                        min_value_length=None,
                    ),
                ),
                (
                    'translated_slug',
                    bx_django_utils.translation.TranslationSlugField(
                        additional_uniqueness=None,
                        blank=True,
                        languages=(('de-de', 'German'), ('en-us', 'US-English'), ('es', 'Spanish')),
                        max_value_length=None,
                        min_value_length=None,
                        unique=False,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='PolymorphicVehicle',
            fields=[
                ('license_plate', models.CharField(max_length=64, primary_key=True, serialize=False)),
                ('color', models.TextField()),
                (
                    'polymorphic_ctype',
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='polymorphic_%(app_label)s.%(class)s_set+',
                        to='contenttypes.contenttype',
                    ),
                ),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
        ),
        migrations.CreateModel(
            name='StoreSaveModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name='TimetrackingTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'create_dt',
                    models.DateTimeField(
                        blank=True,
                        editable=False,
                        help_text='ModelTimetrackingMixin.create_dt.help_text',
                        null=True,
                        verbose_name='ModelTimetrackingMixin.create_dt.verbose_name',
                    ),
                ),
                (
                    'update_dt',
                    models.DateTimeField(
                        blank=True,
                        editable=False,
                        help_text='ModelTimetrackingMixin.update_dt.help_text',
                        null=True,
                        verbose_name='ModelTimetrackingMixin.update_dt.verbose_name',
                    ),
                ),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TranslatedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'translated',
                    bx_django_utils.translation.TranslationField(
                        languages=(('de-de', 'de-de'), ('en-us', 'en-us'), ('es', 'es')),
                        max_value_length=None,
                        min_value_length=None,
                    ),
                ),
                (
                    'translated_multiline',
                    bx_django_utils.translation.TranslationField(
                        blank=True,
                        languages=(('de-de', 'German'), ('en-us', 'US-English'), ('es', 'Spanish')),
                        max_value_length=None,
                        min_value_length=None,
                    ),
                ),
                ('not_translated', models.TextField(default='A Default Value')),
            ],
        ),
        migrations.CreateModel(
            name='TranslatedSlugTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'translated',
                    bx_django_utils.translation.TranslationField(
                        languages=(('de-de', 'German'), ('en-us', 'US-English'), ('es', 'Spanish')),
                        max_value_length=None,
                        min_value_length=None,
                    ),
                ),
                (
                    'translated_slug',
                    bx_django_utils.translation.TranslationSlugField(
                        additional_uniqueness=None,
                        blank=True,
                        languages=(('de-de', 'de-de'), ('en-us', 'en-us'), ('es', 'es')),
                        max_value_length=None,
                        min_value_length=None,
                        unique=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='ValidateLengthTranslations',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'translated',
                    bx_django_utils.translation.TranslationField(
                        languages=(('de', 'German'), ('en', 'English')), max_value_length=20, min_value_length=3
                    ),
                ),
                (
                    'translated_slug',
                    bx_django_utils.translation.TranslationSlugField(
                        additional_uniqueness=None,
                        blank=True,
                        languages=(('de', 'German'), ('en', 'English')),
                        max_value_length=20,
                        min_value_length=3,
                        unique=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='PolymorphicBike',
            fields=[
                (
                    'polymorphicvehicle_ptr',
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to='test_app.polymorphicvehicle',
                    ),
                ),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('test_app.polymorphicvehicle',),
        ),
        migrations.CreateModel(
            name='PolymorphicCar',
            fields=[
                (
                    'polymorphicvehicle_ptr',
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to='test_app.polymorphicvehicle',
                    ),
                ),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('test_app.polymorphicvehicle',),
        ),
        migrations.CreateModel(
            name='CreateOrUpdateTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'create_dt',
                    models.DateTimeField(
                        blank=True,
                        editable=False,
                        help_text='ModelTimetrackingMixin.create_dt.help_text',
                        null=True,
                        verbose_name='ModelTimetrackingMixin.create_dt.verbose_name',
                    ),
                ),
                (
                    'update_dt',
                    models.DateTimeField(
                        blank=True,
                        editable=False,
                        help_text='ModelTimetrackingMixin.update_dt.help_text',
                        null=True,
                        verbose_name='ModelTimetrackingMixin.update_dt.verbose_name',
                    ),
                ),
                ('name', models.CharField(max_length=64)),
                ('slug', models.SlugField(max_length=64)),
                ('blank_field', models.CharField(blank=True, max_length=64)),
                ('null_field', models.CharField(blank=True, max_length=64, null=True)),
                ('uuid_field', models.UUIDField(blank=True, null=True)),
                (
                    'many2one_rel',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to='test_app.timetrackingtestmodel',
                    ),
                ),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import datetime
from unittest import mock
from unittest.mock import patch

from bx_py_utils.test_utils.datetime import parse_dt
from django.contrib.admin.models import LogEntry
from django.core.cache import cache
from django.template.defaulttags import CsrfTokenNode
from django.test import TestCase
from django.utils import timezone

from bx_django_utils.feature_flags.test_utils import (
    FeatureFlagTestCaseMixin,
    get_feature_flag_cache_info,
    get_feature_flag_db_info,
    get_feature_flag_states,
)
from bx_django_utils.templatetags.humanize_time import human_duration
from bx_django_utils.test_utils.datetime import MockDatetimeGenerator
from bx_django_utils.test_utils.html_assertion import (
    HtmlAssertionMixin,
    assert_html_response_snapshot,
    get_django_name_suffix,
)
from bx_django_utils.test_utils.users import make_test_user
from bx_django_utils_tests.test_app.feature_flags import bar_feature_flag, foo_feature_flag


class FeatureFlagIntegrationTestCase(FeatureFlagTestCaseMixin, HtmlAssertionMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.superuser = make_test_user(username='test-superuser', is_superuser=True)

    @patch.object(timezone, 'now', MockDatetimeGenerator(datetime.timedelta(minutes=1)))
    def test_basic(self):
        self.client.force_login(self.superuser)
        response = self.client.get('/admin/')
        self.assert_html_parts(
            response,
            parts=(
                '<a href="/admin/feature_flags/manage/">Manage Feature Flags</a>',
                (
                    '<a href="/admin/feature_flags/feature-flags-values-demo/">'
                    'Feature Flags values DEMO</a>'
                ),
            ),
        )

        with mock.patch.object(CsrfTokenNode, 'render', return_value='MockedCsrfTokenNode'):
            response = self.client.get('/admin/feature_flags/manage/')
        self.assert_html_parts(
            response,
            parts=(
                '<h2 id="foo">Foo - ENABLED&nbsp;<a href="#foo">&para;</a></h2>',
                '<input type="submit" value="Set \'Foo\' to DISABLED">',
                '<h2 id="bar">Bar - DISABLED&nbsp;<a href="#bar">&para;</a></h2>',
                '<input type="submit" value="Set \'Bar\' to ENABLED">',
            ),
        )
        assert_html_response_snapshot(
            response,
            query_selector='#content',
            validate=False,
            name_suffix=get_django_name_suffix(),
        )

        self.assertEqual(foo_feature_flag.state.name, 'ENABLED')
        self.assertEqual(bar_feature_flag.state.name, 'DISABLED')
        self.assert_messages(response, expected_messages=[])

        response = self.client.get('/admin/feature_flags/feature-flags-values-demo/')
        self.assertRedirects(response, expected_url='/admin/', fetch_redirect_response=False)
        self.assert_messages(
            response,
            expected_messages=[
                'Demo Feature "Foo" state: ENABLED',
                'Demo Feature "Bar" state: DISABLED',
            ],
        )

        self.client = self.get_fresh_client()  # "Clear" messages by recreate test client

        response = self.client.post(
            '/admin/feature_flags/manage/',
            data={
                'cache_key': 'feature-flags-foo',
                'new_value': '1',
            },
        )
        self.assert_messages(response, expected_messages=['Current "Foo" state is already: ENABLED'])
        self.assertRedirects(
            response,
            expected_url='/admin/feature_flags/manage/',
            fetch_redirect_response=False,
            msg_prefix=response.content.decode('utf-8'),
        )
        self.assertEqual(LogEntry.objects.count(), 0)

        self.client = self.get_fresh_client()  # "Clear" messages by recreate test client

        response = self.client.post(
            '/admin/feature_flags/manage/',
            data={
                'cache_key': 'feature-flags-foo',
                'new_value': '0',
            },
        )
        self.assert_messages(response, expected_messages=['Set "Foo" to DISABLED'])
        self.assertRedirects(
            response,
            expected_url='/admin/feature_flags/manage/',
            fetch_redirect_response=False,
            msg_prefix=response.content.decode('utf-8'),
        )
        log_entry = LogEntry.objects.get()
        self.assertEqual(log_entry.change_message, 'Changed feature flag "Foo"')
        self.assertEqual(log_entry.object_repr, 'Set "Foo" to DISABLED')
        self.assertEqual(log_entry.object_id, 'feature-flags-foo')

        self.assertTrue(log_entry.action_time)  # Time is set?

        # "Mock" the real time:
        action_time = parse_dt('2000-06-15T00:00:00+0000')
        LogEntry.objects.update(action_time=action_time)

        with (patch.object(CsrfTokenNode, 'render', return_value='MockedCsrfTokenNode'),):
            response = self.client.get('/admin/feature_flags/manage/')
        self.assert_html_parts(
            response,
            parts=(
                '<h2 id="foo">Foo - DISABLED&nbsp;<a href="#foo">&para;</a></h2>',
                '<h2 id="bar">Bar - DISABLED&nbsp;<a href="#bar">&para;</a></h2>',
                f'<p>Last changed by test-superuser {human_duration(action_time)} ago.</p>',
            ),
        )
        assert_html_response_snapshot(response, query_selector='#content', validate=False)


class PersistentFeatureFlagTestCase(FeatureFlagTestCaseMixin, TestCase):
    warum_up_feature_flag_cache = False

    def test_persistent(self):
        # Nothing stored, yet:
        self.assertEqual(get_feature_flag_cache_info(), {})
        self.assertEqual(get_feature_flag_db_info(), {})

        # Check initial state of "Foo"-Flag:
        self.assertTrue(foo_feature_flag.is_enabled)  # Foo initial == enabled

        # Check will persistent the current stage of the flag:
        self.assertEqual(get_feature_flag_cache_info(), {'feature-flags-foo': 1})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-foo': 1})

        # Check initial state of "Bar"-Flag:
        self.assertFalse(bar_feature_flag.is_enabled)  # Bar initial == disabled

        # Now both are persistent:
        self.assertEqual(get_feature_flag_cache_info(), {'feature-flags-bar': 0, 'feature-flags-foo': 1})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-bar': 0, 'feature-flags-foo': 1})
        # Check get_feature_flag_states():
        self.assertEqual(get_feature_flag_states(), {'feature-flags-bar': False, 'feature-flags-foo': True})

        # toggle flags:
        foo_feature_flag.disable()
        self.assertEqual(get_feature_flag_states(), {'feature-flags-bar': False, 'feature-flags-foo': False})
        bar_feature_flag.enable()
        self.assertEqual(get_feature_flag_states(), {'feature-flags-bar': True, 'feature-flags-foo': False})

        # Check state:
        self.assertFalse(foo_feature_flag.is_enabled)
        self.assertTrue(bar_feature_flag.is_enabled)

        # Remove the states from the cache:
        cache.clear()

        # Cache empty, but database holds still the data:
        self.assertEqual(get_feature_flag_cache_info(), {})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-bar': 1, 'feature-flags-foo': 0})

        # State should be not changed:
        self.assertFalse(foo_feature_flag.is_enabled)
        self.assertTrue(bar_feature_flag.is_enabled)

        # Now it's in the cache:
        self.assertEqual(get_feature_flag_cache_info(), {'feature-flags-bar': 1, 'feature-flags-foo': 0})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-bar': 1, 'feature-flags-foo': 0})

    def test_temp_enable(self):
        self.assertTrue(foo_feature_flag)
        self.assertFalse(bar_feature_flag)

        with (
            foo_feature_flag.temp_enable(),
            bar_feature_flag.temp_enable(),
        ):
            self.assertTrue(foo_feature_flag)
            self.assertTrue(bar_feature_flag)

        self.assertTrue(foo_feature_flag)
        self.assertFalse(bar_feature_flag)

    def test_temp_disable(self):
        self.assertTrue(foo_feature_flag)
        self.assertFalse(bar_feature_flag)

        with (
            foo_feature_flag.temp_disable(),
            bar_feature_flag.temp_disable(),
        ):
            self.assertFalse(foo_feature_flag)
            self.assertFalse(bar_feature_flag)

        self.assertTrue(foo_feature_flag)
        self.assertFalse(bar_feature_flag)



from django.contrib.admin.models import ACTION_FLAG_CHOICES, LogEntry
from django.contrib.auth.models import User
from django.test import TestCase
from model_bakery import baker

from bx_django_utils.admin_extra_views.utils import reverse_admin_extra_view
from bx_django_utils.test_utils.html_assertion import HtmlAssertionMixin
from bx_django_utils_tests.test_app.admin_views import GenericModelFilterAdminExtraView


class GenericModelFilterTestCase(HtmlAssertionMixin, TestCase):
    """[no-doc]"""

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.super_user = User.objects.create_superuser(username='foobar')

    def test_basics(self):
        url = reverse_admin_extra_view(GenericModelFilterAdminExtraView)
        self.assertEqual(url, '/admin/generic-model-filter/generic-filter/')

        for action_choice in ACTION_FLAG_CHOICES:
            action = action_choice[0]
            for object_id in (None, 'foo', 'foo', 'bar'):
                baker.make(LogEntry, object_id=object_id, action_flag=action)

        self.client.force_login(self.super_user)

        #########################################################################
        # Step 1: selected model to be searched

        response = self.client.get('/admin/generic-model-filter/generic-filter/')
        self.assertTemplateUsed(response, template_name='generic_model_filter.html')
        self.assert_html_parts(
            response,
            parts=(
                '<h1>Generic model item filter</h1>',
                '<h2>Step 1: selected model to be searched</h2>',
                '<label for="id_model_name">Model Name:</label>',
                '<option value="auth.User">Authentication and Authorization | user</option>',
            ),
        )

        response = self.client.post(
            path='/admin/generic-model-filter/generic-filter/', data={'model_name': 'admin.LogEntry'}
        )
        self.assertRedirects(
            response,
            expected_url='/admin/generic-model-filter/generic-filter/?model_name=admin.LogEntry',
            fetch_redirect_response=False,
        )

        #########################################################################
        # Step 2: selected fields to be searched on the selected model

        response = self.client.get('/admin/generic-model-filter/generic-filter/?model_name=admin.LogEntry')
        self.assert_html_parts(
            response,
            parts=(
                '<h1>Generic model item filter</h1>',
                '<h2>Step 2: selected fields to be searched on the selected model</h2>',
                '<label for="id_field_names">Fields:</label>',
                '<option value="action_flag">action flag</option>',
            ),
        )

        response = self.client.post(
            path='/admin/generic-model-filter/generic-filter/?model_name=admin.LogEntry',
            data={'field_names': ['object_id', 'action_flag']},
        )
        self.assertRedirects(
            response,
            expected_url=(
                '/admin/generic-model-filter/generic-filter/'
                '?model_name=admin.LogEntry'
                '&field_names=object_id'
                '&field_names=action_flag'
            ),
            fetch_redirect_response=False,
        )

        #########################################################################
        # Step 3: The field values used to filter the model

        response = self.client.get(
            path=(
                '/admin/generic-model-filter/generic-filter/'
                '?model_name=admin.LogEntry'
                '&field_names=object_id'
                '&field_names=action_flag'
            )
        )
        self.assert_html_parts(
            response,
            parts=(
                '<h1>Generic model item filter</h1>',
                '<h2>Step 3: The field values used to filter the model</h2>',
                '<label for="id_object_id">Object id:</label>',
                '<option value="1">Addition</option>',
                '<option value="2">Change</option>',
            ),
        )

        self.assertNotIn('generic_model_filter', self.client.session)

        response = self.client.post(
            path=(
                '/admin/generic-model-filter/generic-filter/'
                '?model_name=admin.LogEntry'
                '&field_names=object_id'
                '&field_names=action_flag'
            ),
            data={
                'object_id': 'foo',
                'is_null_check_action_flag': '1',
            },
        )
        self.assertRedirects(
            response,
            expected_url=(
                '/admin/generic-model-filter/generic-filter/'
                '?model_name=admin.LogEntry'
                '&field_names=object_id'
                '&field_names=action_flag'
            ),
            fetch_redirect_response=False,
        )
        self.assertIn('generic_model_filter', self.client.session)
        data = self.client.session['generic_model_filter']
        self.assertEqual(
            data,
            {'object_id': 'foo', 'action_flag': '', 'is_null_check_object_id': '', 'is_null_check_action_flag': '1'},
        )

        #########################################################################
        # Search result

        response = self.client.get(
            path=(
                '/admin/generic-model-filter/generic-filter/'
                '?model_name=admin.LogEntry'
                '&field_names=object_id'
                '&field_names=action_flag'
            )
        )
        self.assert_html_parts(
            response,
            parts=(
                '<h1>Generic model item filter</h1>',
                '<h2>Found 6 entries (total: 12)</h2>',
                '<span class="current">Page 1 of 1.</span>',
            ),
        )



import functools
from typing import TYPE_CHECKING

from django.core.cache.backends.base import memcache_key_warnings
from django.core.exceptions import ValidationError

from bx_django_utils.feature_flags.exceptions import FeatureFlagDisabled


if TYPE_CHECKING:
    from bx_django_utils.feature_flags.data_classes import FeatureFlag


def validate_cache_key(cache_key: str) -> None:
    messages = list(memcache_key_warnings(cache_key))
    if messages:
        raise ValidationError(messages)


def if_feature(feature_flag: "FeatureFlag", raise_exception: bool = False):
    """
    A decorator that only executed the decorated function if the given feature flag is enabled.
    Performs a noop otherwise, or raises a FeatureFlagDisabled exception if raise_exception is True.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not feature_flag.is_enabled:
                if not raise_exception:
                    return
                else:
                    raise FeatureFlagDisabled(feature_flag)
            return func(*args, **kwargs)

        return wrapper

    return decorator



from django.apps import AppConfig


class GenericModelFilterAppConfig(AppConfig):
    """"""  # noqa - don't add in README

    name = 'bx_django_utils.generic_model_filter'
    verbose_name = 'Generic Model Filter'



from collections.abc import Iterable

from django import forms
from django.apps import AppConfig, apps
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.options import Options


def iter_user_models(user, permission_codename: str = 'view') -> Iterable[models.Model]:
    """
    Filter models for the given user.
    permission_codename e.g.: "view" | "add" | "change" | "delete" etc.
    """
    app_configs = apps.get_app_configs()
    for app_config in app_configs:
        models = app_config.get_models()
        for model in models:
            content_type = ContentType.objects.get_for_model(model)
            permissions = Permission.objects.filter(
                content_type=content_type,
                codename__startswith=permission_codename,
            )
            if user.has_perms(permissions):
                yield model


def get_user_model_choices(user, permission_codename: str = 'view'):
    """
    Build a form choices list with all models the given user can "view" | "add" | "change" | "delete" etc.
    """
    choices = []
    for ModelClass in iter_user_models(user, permission_codename=permission_codename):
        options: Options = ModelClass._meta
        app_config: AppConfig = options.app_config

        verbose_name = f'{app_config.verbose_name} | {options.verbose_name}'
        choices.append((options.label, verbose_name))
    return choices


class SelectModelForm(forms.Form):
    """
    Form to select a model that the user can "view"
    """

    model_name = forms.ChoiceField(
        label='Model Name:',
        help_text='Only entries are visible to which you have read access.',
    )

    def __init__(self, *args, user, **kwargs):
        super().__init__(*args, **kwargs)

        field = self.fields['model_name']
        field.choices = get_user_model_choices(user, permission_codename='view')
